

def test_predicate_gt(measure_result, threshold):
    """
    :param measure_result:
    :param threshold:
    :return:
    """
    if measure_result > threshold:
        return "test_fail"
    else:
        return "test_pass"


def test_predicate_lt(measure_result, threshold):
    if measure_result < threshold:
        return "test_fail"
    else:
        return "test_pass"


def test_predicate_ne(measure_result, threshold):
    if measure_result != threshold:
        return "test_fail"
    else:
        return "test_pass"



import abc

from .log import get_null_logger
from .tests import CustomTestFactory, SQLNullTest, SQLDuplicateTest, SQLSetDuplicateTest, CSVNullTest, \
    CSVDuplicateTest, CSVSetDuplicateTest

from .operations import SQLOperations, SQLSetOperations, CSVOperations, CSVSetOperations, CustomSQLOperations


class DQOperator(object):
    def __init__(self, logger=None):
        self.logger = logger if logger is not None else get_null_logger().info

    @abc.abstractmethod
    def build_test(self): pass


class SQLDQOperator(DQOperator):
    def __init__(
            self,
            conn,
            from_,
            threshold,
            *columns,
            logger=None,
            where=None,
            hard_fail=None,
            use_ansi=True
    ):
        """
        :param conn: Database connection object
        :param from_: Schema and table name of table to be tested
        :param threshold: Numeric or dict to specify the threshold for a given test or collection of tests
        :param columns: The column names entered as comma separated positional arguments
        :param logger: Optional logging function. If None is passed then logged messages will be swallowed
        :param where: Conditional logic to be applied to table specified in `from_`
        :param hard_fail: Boolean or dict to specify if test failure should result in terminating the current process
        :param use_ansi: Boolean to specify if SQL query should be complied to ANSI standards
        """
        super().__init__(logger=logger)
        self.conn = conn
        self.dialect = conn.dialect
        self.threshold = threshold
        self.from_ = from_
        self.columns = columns
        self.where = where
        self.hard_fail = hard_fail
        self.use_ansi = use_ansi

        self.test_results = self.execute()

    @abc.abstractmethod
    def build_test(self): pass

    def execute(self):
        test = self.build_test()
        operator = SQLOperations(self.conn, self.logger, test)
        test_results = operator.execute()
        return test_results


class SQLNullCheckOperator(SQLDQOperator):
    def __init__(
            self,
            conn,
            from_,
            threshold,
            *columns,
            logger=None,
            where=None,
            hard_fail=None,
            use_ansi=True
    ):
        """
        :param conn: Database connection object
        :param from_: Schema and table name of table to be tested
        :param threshold: Numeric or dict to specify the threshold for a given test or collection of tests
        :param columns: The column names entered as comma separated positional arguments
        :param logger: Optional logging function. If None is passed then logged messages will be swallowed
        :param where: Conditional logic to be applied to table specified in `from_`
        :param hard_fail: Boolean or dict to specify if test failure should result in terminating the current process
        :param use_ansi: Boolean to specify if SQL query should be complied to ANSI standards
        """
        super().__init__(conn, from_, threshold, *columns, logger=logger, where=where, hard_fail=hard_fail,
                         use_ansi=use_ansi)

    def build_test(self):
        test = SQLNullTest(self.dialect, self.from_, self.threshold, *self.columns, where=self.where,
                           hard_fail=self.hard_fail, use_ansi=self.use_ansi).factory()
        return test


class SQLDuplicateCheckOperator(SQLDQOperator):
    def __init__(
            self,
            conn,
            from_,
            threshold,
            *columns,
            logger=None,
            where=None,
            hard_fail=None,
            use_ansi=True
    ):
        """
        :param conn: Database connection object
        :param from_: Schema and table name of table to be tested
        :param threshold: Numeric or dict to specify the threshold for a given test or collection of tests
        :param columns: The column names entered as comma separated positional arguments
        :param logger: Optional logging function. If None is passed then logged messages will be swallowed
        :param where: Conditional logic to be applied to table specified in `from_`
        :param hard_fail: Boolean or dict to specify if test failure should result in terminating the current process
        :param use_ansi: Boolean to specify if SQL query should be complied to ANSI standards
        """
        super().__init__(conn, from_, threshold, *columns, logger=logger, where=where, hard_fail=hard_fail,
                         use_ansi=use_ansi)

    def build_test(self):
        test = SQLDuplicateTest(self.dialect, self.from_, self.threshold, *self.columns, where=self.where,
                                hard_fail=self.hard_fail, use_ansi=self.use_ansi).factory()
        return test


class SQLSetDuplicateCheckOperator(SQLDQOperator):
    def __init__(
            self,
            conn,
            from_,
            threshold,
            *columns,
            logger=None,
            where=None,
            hard_fail=None,
            use_ansi=True
    ):
        """
        :param conn: Database connection object
        :param from_: Schema and table name of table to be tested
        :param threshold: Numeric or dict to specify the threshold for a given test or collection of tests
        :param columns: The column names entered as comma separated positional arguments
        :param logger: Optional logging function. If None is passed then logged messages will be swallowed
        :param where: Conditional logic to be applied to table specified in `from_`
        :param hard_fail: Boolean or dict to specify if test failure should result in terminating the current process
        :param use_ansi: Boolean to specify if SQL query should be complied to ANSI standards
        """
        super().__init__(conn, from_, threshold, *columns, logger=logger, where=where, hard_fail=hard_fail,
                         use_ansi=use_ansi)

    def build_test(self):
        test = SQLSetDuplicateTest(self.dialect, self.from_, self.threshold, *self.columns, where=self.where,
                                   hard_fail=self.hard_fail, use_ansi=self.use_ansi).factory()
        return test

    def execute(self):
        test = self.build_test()
        operator = SQLSetOperations(self.conn, self.logger, test)
        test_results = operator.execute()
        return test_results


class CSVDQOperator(DQOperator):
    def __init__(
            self,
            from_,
            threshold,
            *columns,
            delimiter=',',
            hard_fail=None,
            fieldnames=None,
            logger=None,
            **reducer_kwargs
    ):
        """
        :param from_: Path to CSV file to be tested
        :param threshold: Numeric or dict to specify the threshold for a given test or collection of tests
        :param columns: Column names entered as comma separated positional arguments
        :param delimiter: Separator used in file specified by the `from_` parameter
        :param hard_fail: Boolean or dict to specify if test failure should result in terminating the current process
        :param fieldnames: Sequence of all columns in CSV specified by `from_`. To be used when no headers exist in file
        :param logger: Optional logging function. If None is passed then logged messages will be swallowed
        :param reducer_kwargs: Key word arguments passed to the measure reducer function
        """
        super().__init__(logger=logger)
        self.from_ = from_
        self.threshold = threshold
        self.columns = columns
        self.delimiter = delimiter
        self.hard_fail = hard_fail
        self.fieldnames = fieldnames
        self.reducer_kwargs = reducer_kwargs

        self.test_results = self.execute()

    @abc.abstractmethod
    def build_test(self): pass

    def execute(self):
        test = self.build_test()
        operator = CSVOperations(self.logger, test, fieldnames=self.fieldnames, **self.reducer_kwargs)
        test_results = operator.execute()
        return test_results


class CSVNullCheckOperator(CSVDQOperator):
    def __init__(
            self,
            from_,
            threshold,
            *columns,
            delimiter=',',
            hard_fail=None,
            fieldnames=None,
            logger=None

    ):
        """
        :param from_: Path to CSV file to be tested
        :param threshold: Numeric or dict to specify the threshold for a given test or collection of tests
        :param columns: Column names entered as comma separated positional arguments
        :param delimiter: Separator used in file specified by the `from_` parameter
        :param hard_fail: Boolean or dict to specify if test failure should result in terminating the current process
        :param fieldnames: Sequence of all columns in CSV specified by `from_`. To be used when no headers exist in file
        :param logger: Optional logging function. If None is passed then logged messages will be swallowed
        """
        super().__init__(from_, threshold, *columns, delimiter=delimiter, hard_fail=hard_fail, fieldnames=fieldnames,
                         logger=logger)

    def build_test(self):
        test = CSVNullTest(self.from_, self.threshold, *self.columns, delimiter=self.delimiter,
                           hard_fail=self.hard_fail).factory()
        return test


class CSVDuplicateCheckOperator(CSVDQOperator):
    def __init__(
            self,
            from_,
            threshold,
            *columns,
            delimiter=',',
            hard_fail=None,
            fieldnames=None,
            logger=None
    ):
        """
        :param from_: Path to CSV file to be tested
        :param threshold: Numeric or dict to specify the threshold for a given test or collection of tests
        :param columns: Column names entered as comma separated positional arguments
        :param delimiter: Separator used in file specified by the `from_` parameter
        :param hard_fail: Boolean or dict to specify if test failure should result in terminating the current process
        :param fieldnames: Sequence of all columns in CSV specified by `from_`. To be used when no headers exist in file
        :param logger: Optional logging function. If None is passed then logged messages will be swallowed
        """
        super().__init__(from_, threshold, *columns, delimiter=delimiter, hard_fail=hard_fail, fieldnames=fieldnames,
                         logger=logger)

    def build_test(self):
        test = CSVDuplicateTest(self.from_, self.threshold, *self.columns, delimiter=self.delimiter,
                                hard_fail=self.hard_fail).factory()
        return test


class CSVSetDuplicateCheckOperator(CSVDQOperator):
    def __init__(
            self,
            from_,
            threshold,
            *columns,
            delimiter=',',
            hard_fail=None,
            fieldnames=None,
            logger=None
    ):
        """
        :param from_: Path to CSV file to be tested
        :param threshold: Numeric or dict to specify the threshold for a given test or collection of tests
        :param columns: Column names entered as comma separated positional arguments
        :param delimiter: Separator used in file specified by the `from_` parameter
        :param hard_fail: Boolean or dict to specify if test failure should result in terminating the current process
        :param fieldnames: Sequence of all columns in CSV specified by `from_`. To be used when no headers exist in file
        :param logger: Optional logging function. If None is passed then logged messages will be swallowed
        """
        super().__init__(from_, threshold, *columns, delimiter=delimiter, hard_fail=hard_fail, fieldnames=fieldnames,
                         logger=logger)

    def build_test(self):
        test = CSVSetDuplicateTest(self.from_, self.threshold, *self.columns, delimiter=self.delimiter,
                                   hard_fail=self.hard_fail).factory()
        return test

    def execute(self):
        test = self.build_test()
        operator = CSVSetOperations(self.logger, test, fieldnames=self.fieldnames, **self.reducer_kwargs)
        test_results = operator.execute()
        return test_results


class CustomSQLDQOperator(DQOperator):
    def __init__(
            self,
            conn,
            custom_test,
            description,
            *columns,
            threshold=None,
            hard_fail=None,
            logger=None,
            **test_desc_kwargs
    ):
        """
        :param conn: Database connection object
        :param custom_test: SQL query used to execute the data quality test
        :param description: Description of the test given in `custom_test`
        :param columns: Column names entered as comma separated positional arguments
        :param threshold: Numeric or dict to specify the threshold for a given test or collection of tests
        :param hard_fail: Boolean or dict to specify if test failure should result in terminating the current process
        :param logger: Optional logging function. If None is passed then logged messages will be swallowed
        :param test_desc_kwargs: Key word arguments passed to the test description formatter
        """
        super().__init__(logger=logger)
        self.conn = conn
        self.description = description
        self.custom_test = custom_test
        self.columns = columns
        self.threshold = threshold
        self.hard_fail = hard_fail
        self.test_desc_kwargs = test_desc_kwargs

        self.test_results = self.execute()

    def build_test(self):
        test = CustomTestFactory(self.description, self.custom_test, *self.columns, threshold=self.threshold,
                                 hard_fail=self.hard_fail).factory()
        return test

    def execute(self):
        test = self.build_test()
        operator = CustomSQLOperations(self.conn, self.logger, test, **self.test_desc_kwargs)
        test_results = operator.execute()
        return test_results



import sqlalchemy.sql.expression as sql
from sqlalchemy.sql import func
from sqlalchemy import text, Float

from .core import build_select_query, build_aggregate_query


def measure_proportion_each_column(from_clause, aggregate_func, *columns, where_clause=None):
    """
    :param from_clause: used as query from clause; expected sqlalchemy TextClause, Select, Alias or Join object
    :param aggregate_func: the sql aggregation function
    :param columns: column names to be selected; expected str, TextClause or ColumnClause
    :param where_clause: predicate to apply to where clause in query; expected str, TextClause, BinaryExpression
    :return: sqlalchemy.sql.selectable.Select object
    """
    count_rows_aggregate = {"1": func.count}
    column_aggregates = {column: aggregate_func for column in columns}
    aggregates = {**count_rows_aggregate, **column_aggregates}

    aggregate_query = build_aggregate_query(from_clause, where_clause=where_clause, **aggregates)
    aggregate_query = aggregate_query.alias('t')
    aggregate_query_columns = list(aggregate_query.columns)
    rows_column = aggregate_query_columns[0]
    null_cnt_columns = aggregate_query_columns[1:]

    case_clauses = []
    for i in range(len(columns)):
        column = null_cnt_columns[i]
        column_label = columns[i]
        case_statement = sql.case(
            [
                (
                    rows_column > text("0"),
                    text("1") - sql.cast(column, Float()) / rows_column
                )
            ],
            else_=None
        ).label(column_label)
        case_clauses.append(case_statement)

    measure_query = sql.select(case_clauses).select_from(aggregate_query)
    return measure_query


def measure_set_duplication(from_clause, *columns, where_clause=None):
    """
    :param from_clause: used as query from clause; expected sqlalchemy TextClause, Select, Alias or Join object
    :param columns: column names to be selected; expected str, TextClause or ColumnClause
    :param where_clause: predicate to apply to where clause in query; expected str, TextClause, BinaryExpression
    :return: sqlalchemy.sql.selectable.Select object
    """
    count_rows_aggregate = {"1": func.count}

    rows_query = build_aggregate_query(from_clause, where_clause=where_clause, **count_rows_aggregate)
    rows_query = rows_query.alias('r')
    rows_column = list(rows_query.columns)[0]

    distinct_query = build_select_query(from_clause, *columns, where_clause=where_clause, select_distinct=True)
    distinct_query = distinct_query.alias('t')

    uniques_query = build_aggregate_query(distinct_query, **count_rows_aggregate)
    uniques_query = uniques_query.alias('u')
    uniques_column = list(uniques_query.columns)[0]

    join_clause = sql.join(rows_query, uniques_query, onclause=rows_column != None)

    measure_label = ",".join(columns)
    measure_query = sql.select(
        [
            sql.case(
                [
                    (
                        rows_column > text("0"),
                        text("1") - sql.cast(uniques_column, Float()) / rows_column
                    )
                ],
                else_=None
            ).label(measure_label)
        ]
    ).select_from(join_clause)
    return measure_query



import csv


def get_csv_document(path, delimiter=',', fieldnames=None):
    """
    :param path: path to csv file
    :param delimiter: separator used in csv file
    :return: list of tuples where each tuple is a row in csv file
    """
    with open(path, 'r') as infile:
        csvreader = csv.DictReader(infile, delimiter=delimiter, fieldnames=fieldnames)
        dataset = list(csvreader)
    infile.close()
    return dataset


def apply_reducer(dataset, reducer, *columns, **kwargs):
    """
    :param dataset:
    :param measure:
    :param accum:
    :param args:
    :param is_header_included:
    :param kwargs:
    :return:
    """
    rowcnt = 0
    accum = {}
    for row in dataset:
        rowcnt += 1
        output = reducer(row, *columns, **kwargs)

        result = output["result"]
        collection = output.get("collection")
        if collection is not None:
            kwargs["collection"] = collection

        for column in result:
            accum[column] = accum.get(column, 0) + result[column]

    results = {"rowcnt": rowcnt, "accum": accum}
    return results



import sqlalchemy as db
from sqlalchemy import text

from .exception_handling import try_except
from .log import get_null_logger


class DBConnector(object):
    def __init__(self, user, password, host, dbname, port, logger=None):
        """
        :param user: user name
        :param password: user password
        :param host: database host
        :param dbname: database name
        :param port: database port
        :param logger: optional logging function
        """
        self.user = user
        self.password = password
        self.host = host
        self.dbname = dbname
        self.port = port
        self.conn = None
        self.logger = logger if logger is not None else get_null_logger().debug

    def __get_credential(self):
        pass

    def __get_engine(self, credential):
        """
        :param credential: database login crediential
        :return: sqlalchemy.engine.base.Engine object
        """
        engine = db.create_engine(credential)
        return engine

    def get_conn(self, credential):
        """
        :param credential: database login crediential
        :return: sqlalchemy.engine.base.Connection object
        """
        @try_except(self.logger)
        def apply():
            engine = self.__get_engine(credential)
            conn = engine.connect()
            return conn
        return apply()

    def execute(self, query):
        """
        :param query: SQL query to be executed
        :return: sqlalchemy.engine.result.ResultProxy object
        """
        @try_except(self.logger)
        def apply():
            if isinstance(query, str):
                query_ = text(query)
            else:
                query_ = query
            response = self.conn.execute(query_)
            return response

        if self.conn is not None:
            return apply()
        else:
            raise AttributeError("expected sqlalchemy.engine.connect object but found None")

    def fetch(self, response):
        """
        :param response: sqlalchemy.engine.result.ResultProxy object
        :return: list of tuples where each tuple is a row from the query result
        """
        query_result = response.fetchall()
        return query_result


class PostgresConnector(DBConnector):
    def __init__(self, user, password, host, dbname, port, logger=None):
        super().__init__(user, password, host, dbname, port, logger=logger)
        credential = self.__get_credential()
        self.conn = self.get_conn(credential)
        self.dialect = "POSTGRES"

    def __get_credential(self):
        cred = f"postgresql://{self.user}:{self.password}@{self.host}:{self.port}/{self.dbname}"
        return cred


class MySQLConnector(DBConnector):
    def __init__(self, user, password, host, dbname, port, logger=None):
        super().__init__(user, password, host, dbname, port, logger=logger)
        credential = self.__get_credential()
        self.conn = self.get_conn(credential)
        self.dialect = "MYSQL"

    def __get_credential(self):
        cred = f"mysql+pymysql://{self.user}:{self.password}@{self.host}:{self.port}/{self.dbname}"
        return cred




def measure_null(row, *columns, null_values=None):
    """
    :param row:
    :param column_indicies:
    :param null_values:
    :return:
    """
    null_values_ = {"NULL"}
    if null_values is not None:
        null_values_ = null_values_.union(null_values)

    result = {}
    for column in columns:
        value = row[column]
        if value in null_values_:
            result[column] = 1
        else:
            result[column] = 0
    output = {"result": result}
    return output


def measure_duplicates(row, *columns, collection=None):
    """
    :param row:
    :param column_indicies:
    :param collection:
    :return:
    """
    result = {}
    if collection is None:
        collection = {}

    for column in columns:
        value = row[column]
        key = (column, value)
        if key not in collection:
            collection[key] = None
            result[column] = 0
        else:
            result[column] = 1

    output = {"result": result, "collection": collection}
    return output


def measure_set_duplicates(row, *columns, collection=None):
    """
    :param row:
    :param columns:
    :param collection:
    :return:
    """
    if collection is None:
        collection = {}

    values = tuple(map(lambda col: row[col], columns))
    key = tuple(zip(columns, values))

    columns_label = ",".join(columns)
    if key not in collection:
        collection[key] = None
        result = {columns_label: 0}
    else:
        result = {columns_label: 1}
    output = {"result": result, "collection": collection}
    return output





def load_template(path):
    with open(path, 'r') as infile:
        template = infile.read()
    infile.close()
    return template


def test_reuslt_msg_template():
    path = "dataraven/static/test_result_message_template.txt"
    return load_template(path)


def hard_fail_msg_template():
    path = "dataraven/static/hard_fail_message_template.txt"
    return load_template(path)


import traceback
from functools import wraps
from datetime import datetime


class TestFailure(Exception):
    def __init__(self, message):
        super().__init__(message)


def try_except(logger, msg=None):
    """
    :param logger: logging function
    :param msg: optional exception message. should contain place holders for: current timestamp, traceback
    :return: function call contained in wrapper or raise appropriate exception
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as ex:
                now = datetime.strftime(datetime.now(), "%Y-%m-%d %H:%M:%S")
                tb = traceback.format_exc()
                if msg:
                    error_msg = msg.format(now=now, tb=tb)
                else:
                    error_msg = f"caught exception while calling {func.__name__} at {now}. traceback: {tb}"
                logger(error_msg)
                raise ex
        return wrapper
    return decorator



from sqlalchemy.dialects.postgresql import dialect as PostgresDialect
from sqlalchemy.dialects.mysql import dialect as MySQLDialect
from sqlalchemy import text, column
from sqlalchemy.sql.elements import TextClause, ColumnClause
from sqlalchemy.sql.selectable import Select, Alias, Join


def apply_where_clause(query, where_clause):
    """
    :param query: sqlalchemy.sql.selectable.Select object
    :param where_clause: predicate to apply to where clause in query; expected str, TextClause, BinaryExpression
    :return: query with where clause logic added
    """
    if isinstance(where_clause, str):
        where_clause_ = text(where_clause)
        query_ = query.where(where_clause_)
        return query_
    else:
        try:
            query_ = query.where(where_clause)
            return query_
        except TypeError:
            type_ = type(where_clause)
            error_msg = f"{apply_where_clause.__name__}, caught unexpected type for where_clause: {type_}"
            raise TypeError(error_msg)
        except Exception as ex:
            raise ex


def format_from_clause(from_clause):
    """
    :param from_clause: used as query from clause; expected sqlalchemy TextClause, Select, Alias or Join object
    :return: return formatted from clause
    """
    expected_sqlalchemy_classes = [TextClause, Select, Alias, Join]
    condition = any(map(lambda class_: isinstance(from_clause, class_), expected_sqlalchemy_classes))

    if isinstance(from_clause, str):
        return text(from_clause)
    elif condition is True:
        return from_clause
    else:
        expected_sqlalchemy_classes_ = ','.join(list(map(lambda class_: str(class_.__name__),
                                                         expected_sqlalchemy_classes)))
        type_ = type(from_clause)
        error_msg = f"expected from_clause to be type: {expected_sqlalchemy_classes_} but found {type_}"
        raise TypeError(error_msg)


def format_select_columns(*columns):
    """
    :param columns: column names to be selected; expected str, TextClause or ColumnClause
    :return: tuple of column names
    """
    output = []
    for col in columns:
        if isinstance(col, str):
            output.append(column(col))
        elif isinstance(col, TextClause):
            output.append(column(str(col)))
        elif isinstance(col, ColumnClause):
            output.append(col)
        else:
            type_ = type(col)
            error_msg = f"expected column to be of type str, TextClause or ColumnClause but found {type_}"
            raise TypeError(error_msg)
    return tuple(output)


def compile_to_dialect(query, dialect, use_ansi=True):
    """
    :param query: sqlalchemy.sql.selectable.Select object
    :param dialect: str to be used as key to get correct dialect
    :param use_ansi: compile to ansi compliant sql if True
    :return: formatted query to specified dialect
    """
    dialect_ = dialect.upper()
    if dialect_ == "POSTGRES":
        query_ = query.compile(dialect=PostgresDialect(use_ansi=use_ansi))
    elif dialect_ == "MYSQL":
        query_ = query.compile(dialect=MySQLDialect(use_ansi=use_ansi))
    else:
        error_msg = f"expected one fo the following dialects: POSTGRES, MYSQL ... but found {dialect}"
        raise ValueError(error_msg)
    return query_








from collections.abc import Iterable

import sqlalchemy.sql.expression as sql
from sqlalchemy import text

from .helpers import format_select_columns, format_from_clause, apply_where_clause


def build_select_query(from_clause, *columns, where_clause=None, select_distinct=False):
    """
    :param from_clause: sqlalchemy TextClause, Select, Alias or Join object
    :param columns: column names to be selected; expected str, TextClause or ColumnClause
    :param where_clause: predicate to apply to where clause in query; expected str, TextClause, BinaryExpression
    :param select_distinct: change SELECT to SELECT DISTINCT
    :return: sqlalchemy.sql.selectable.Select object
    """
    columns_ = format_select_columns(*columns)
    query = sql.select(columns_)

    if select_distinct is True:
        query = query.distinct()

    from_clause_ = format_from_clause(from_clause)
    query = query.select_from(from_clause_)

    if isinstance(where_clause, Iterable):
        for predicate in where_clause:
            query = apply_where_clause(query, predicate)
    elif where_clause is not None:
        query = apply_where_clause(query, where_clause)

    return query


def build_aggregate_query(from_clause, *groupby_columns, where_clause=None, **aggregate_columns):
    """
    :param from_clause: sqlalchemy TextClause, Select, Alias or Join object
    :param groupby_columns: column names to be used in group by clause; expected str, TextClause or ColumnClause
    :param where_clause: predicate to apply to where clause in query; expected str, TextClause, BinaryExpression
    :param aggregate_columns: key value pairs of the form column_name=func where column_name is the name of table
    column and func is an aggregate function from sqlalchemy.sql.func
    :return: sqlalchemy.sql.selectable.Select object
    """
    from_clause_ = format_from_clause(from_clause)
    aggregates = []
    for col in aggregate_columns:
        col_ = text(col)
        func = aggregate_columns[col]
        agg_column_label = col
        agg_column = func(col_).label(agg_column_label)
        aggregates.append(agg_column)

    groupby_columns_ = format_select_columns(*groupby_columns)

    query_select_clause = list(groupby_columns_)
    query_select_clause.extend(aggregates)

    query = sql.select(query_select_clause).select_from(from_clause_)

    if isinstance(where_clause, Iterable):
        for predicate in where_clause:
            query = apply_where_clause(query, predicate)
    elif where_clause is not None:
        query = apply_where_clause(query, where_clause)

    query = query.group_by(*groupby_columns_)
    return query



import abc

from .exception_handling import TestFailure
from .common import test_reuslt_msg_template, hard_fail_msg_template

from .sql.operations import FetchQueryResults
from .csv.operations import get_csv_document, apply_reducer


class Operations(object):
    def __init__(self, logger, test):
        """
        :param logger: Logging function
        :param test: tests.Test object
        """
        self.logger = logger
        self.test = test

    @staticmethod
    def parse_dict_param(param, column):
        """
        :param param: Data quality operator input parameter
        :param column: The name of a specified columnm
        :return: value
        """
        if isinstance(param, dict):
            value = param.get(column)
        else:
            value = param
        return value

    @staticmethod
    def format_test_result_msgs(test_outcomes, test_descriptions):
        """
        :param test_outcomes: Dict of test outcomes from build_test_outcomes
        :param test_descriptions: Dict of test descriptions from format_test_description
        :return: test_outcomes dict with the added test result message
        """
        test_outcomes_ = test_outcomes.copy()
        for column in test_outcomes_:
            description = test_descriptions.get(column)
            if description is None:
                description = test_descriptions.get("no_column")
            test_outcome = test_outcomes_[column]
            result = test_outcome["result"]
            measure = test_outcome["measure"]
            threshold = test_outcome["threshold"]

            result_template = test_reuslt_msg_template()
            result_message = result_template.format(description=description, result=result, measure=measure,
                                                    threshold=threshold)
            test_outcome["result_msg"] = result_message
        return test_outcomes_

    def build_test_outcomes(self, measure_values):
        """
        :param measure_values: Dict of the form {COLUMN NAME: MEASURE VALUE}
        :return: Dict of the form
        {
            COLUMN NAME: {
                "result": 'test_fail' or 'test_pass',
                "measure": MEASURE VALUE,
                "threshold": TEST THRESHOLD
            }
        }
        """
        test_outcomes = {}
        threshold = self.test.threshold
        predicate = self.test.predicate
        for column in measure_values:
            threshold_ = self.parse_dict_param(threshold, column)
            measure_value = measure_values[column]
            test_result = predicate(measure_value, threshold_)
            test_outcomes[column] = {"result": test_result, "measure": measure_value, "threshold": threshold_}
        return test_outcomes

    def log_test_results(self, test_results):
        """
        :param test_results: Dict generated by format_test_result_msgs
        :return: True if successful
        """
        for column in test_results:
            test_result = test_results[column]
            result_msg = test_result["result_msg"]
            self.logger(result_msg)
        return True

    def raise_execpetion_if_fail(self, test_results):
        """
        :param test_results: Dict generated by format_test_result_msgs
        :return: True if successful
        """
        hard_fail = self.test.hard_fail
        for column in test_results:
            test_result = test_results[column]
            hard_fail_ = self.parse_dict_param(hard_fail, column)
            if hard_fail_ is True:
                outcome = test_result["result"]
                if outcome == "test_fail":
                    result_msg = test_result["result_msg"]
                    error_msg_template = hard_fail_msg_template()
                    error_msg = error_msg_template.format(result_msg=result_msg)
                    raise TestFailure(error_msg)
        return True

    @abc.abstractmethod
    def format_test_description(self, *args, **kwargs): pass

    @abc.abstractmethod
    def calculate_measure_values(self): pass

    def execute(self):
        descriptions = self.format_test_description()

        measure_values = self.calculate_measure_values()
        test_outcomes = self.build_test_outcomes(measure_values)
        test_results = self.format_test_result_msgs(test_outcomes, descriptions)

        self.log_test_results(test_results)
        self.raise_execpetion_if_fail(test_results)

        return test_results


class SQLOperations(Operations):
    def __init__(self, conn, logger, test):
        """
        :param conn: Database connection object
        :param logger: Logging function
        :param test: tests.Test object
        """
        super().__init__(logger, test)
        self.conn = conn

    def format_test_description(self, **description_kwargs):
        descriptions = {}
        description_template = self.test.description
        threshold = self.test.threshold
        measure = self.test.measure
        columns = measure.columns
        from_ = measure.from_
        description_kwargs["from_"] = from_
        for column in columns:
            threshold_ = self.parse_dict_param(threshold, column)
            description_kwargs["threshold"] = threshold_
            description_kwargs["column"] = column
            description = description_template.format(**description_kwargs)
            descriptions[column] = description
        return descriptions

    def calculate_measure_values(self):
        measure = self.test.measure
        query = measure.query
        measure_values = FetchQueryResults(self.conn, query).get_results()
        return measure_values


class SQLSetOperations(SQLOperations):
    def __init__(self, conn, logger, test):
        super().__init__(conn, logger, test)

    def format_test_description(self, **description_kwargs):
        descriptions = {}
        description_template = self.test.description
        threshold = self.test.threshold
        description_kwargs["threshold"] = threshold
        measure = self.test.measure
        columns = measure.columns
        columns_ = ",".join(columns)
        description_kwargs["column"] = columns_
        from_ = measure.from_
        description_kwargs["from_"] = from_
        description = description_template.format(**description_kwargs)
        descriptions[columns_] = description
        return descriptions


class CSVOperations(Operations):
    def __init__(self, logger, test, fieldnames=None, **reducer_kwargs):
        """
        :param logger: logging function
        :param test: tests.Test object
        :param fieldnames: sequence of all column names for CSV file specified in `from_`
        :param reducer_kwargs: optional key word arguments passed to reducer function
        """
        super().__init__(logger, test)
        self.fieldnames = fieldnames
        self.reducer_kwargs = reducer_kwargs

    def format_test_description(self, **description_kwargs):
        descriptions = {}
        description_template = self.test.description
        threshold = self.test.threshold
        measure = self.test.measure
        columns = measure.columns
        from_ = measure.from_
        description_kwargs["from_"] = from_
        for column in columns:
            threshold_ = self.parse_dict_param(threshold, column)
            description_kwargs["threshold"] = threshold_
            description_kwargs["column"] = column
            description = description_template.format(**description_kwargs)
            descriptions[column] = description
        return descriptions

    def calculate_measure_values(self):
        measure = self.test.measure
        delimiter = measure.delimiter
        path = measure.from_
        reducer = measure.reducer
        columns = measure.columns
        document = get_csv_document(path, delimiter=delimiter, fieldnames=self.fieldnames)
        reducer_results = apply_reducer(document, reducer, *columns, **self.reducer_kwargs)
        measure_values = self.build_measure_proportion_values(reducer_results)
        return measure_values

    @staticmethod
    def build_measure_proportion_values(results):
        """
        :param results: Dict of the form {"rowcnt": NUMBER OF ROWS, "accum": ACCUMULATOR OBJECT} where the ACCUMULATOR
        OBJECT is a dict of the form {COLUMN NAME: ACCUMULATED VALUE}
        :return: Dict of the form {COLUMN NAME: MEASURE VALUE} where MEASURE VALUE is a proportion of the accumulated
        value divided by the NUMBER OF ROWS
        """
        measure_values = {}
        rowcnt = results["rowcnt"]
        if rowcnt == 0:
            raise ValueError(f"rowcnt must be greater than 0.")

        accum = results["accum"]
        for column in accum:
            result = accum[column]
            measure_value = result / rowcnt
            measure_values[column] = measure_value

        return measure_values


class CSVSetOperations(CSVOperations):
    def __init__(self, logger, test, fieldnames=None, **reducer_kwargs):
        super().__init__(logger, test, fieldnames=fieldnames, **reducer_kwargs)

    def format_test_description(self, **description_kwargs):
        descriptions = {}
        description_template = self.test.description
        threshold = self.test.threshold
        description_kwargs["threshold"] = threshold
        measure = self.test.measure
        columns = measure.columns
        columns_ = ",".join(columns)
        description_kwargs["column"] = columns_
        from_ = measure.from_
        description_kwargs["from_"] = from_
        description = description_template.format(**description_kwargs)
        descriptions[columns_] = description
        return descriptions


class CustomSQLOperations(Operations):
    def __init__(self, conn, logger, test, **test_desc_kwargs):
        """
        :param conn: Database connection object
        :param logger: Logging function
        :param test: tests.Test object
        :param test_desc_kwargs: optional test description key word arguments for formatting
        """
        super().__init__(logger, test)
        self.conn = conn
        self.test_desc_kwargs = test_desc_kwargs

    def calculate_measure_values(self): pass

    def format_test_description(self, **kwargs):
        test_descriptions = {}
        description_template = self.test.description
        columns = self.test.columns
        threshold = self.test.threshold

        if columns:
            for column in columns:
                threshold_ = self.parse_dict_param(threshold, column)
                kwargs["threshold"] = threshold_
                kwargs["column"] = column
                description = description_template.format(**kwargs)
                test_descriptions[column] = description
        else:
            description = description_template.format(**kwargs)
            test_descriptions["no_column"] = description

        return test_descriptions

    def calcualte_test_results(self):
        def format_test_outcome(outcome):
            measure = outcome.get("measure")
            threshold = outcome.get("threshold")
            result = outcome["result"]
            return {"result": result, "measure": measure, "threshold": threshold}

        test_outcomes = {}
        query = self.test.test
        columns = self.test.columns
        threshold = self.test.threshold

        if columns:
            for column in columns:
                threshold_ = self.parse_dict_param(threshold, column)
                query_ = query.format(column=column, threshold=threshold_)
                test_outcome = FetchQueryResults(self.conn, query_).get_results()
                test_outcomes[column] = format_test_outcome(test_outcome)
        else:
            test_outcome = FetchQueryResults(self.conn, query).get_results()
            column = test_outcome.get("column")
            if column:
                test_outcomes[column] = format_test_outcome(test_outcome)
            else:
                test_outcomes["no_column"] = format_test_outcome(test_outcome)
        return test_outcomes

    def execute(self):
        descriptions = self.format_test_description(**self.test_desc_kwargs)

        test_outcomes = self.calcualte_test_results()

        test_results = self.format_test_result_msgs(test_outcomes, descriptions)

        self.log_test_results(test_results)
        self.raise_execpetion_if_fail(test_results)

        return test_results



import abc

from .measures import SQLNullMeasure, SQLDuplicateMeasure, SQLSetDuplicateMeasure, CSVNullMeasure, \
    CSVDuplicateMeasure, CSVSetDuplicateMeasure

from .test_logic import test_predicate_gt


class Test(object):
    pass


class SQLTest(Test):
    def __init__(self, description, measure, predicate, threshold, hard_fail=False):
        """
        :param description: description of test
        :param measure: measures.Measure object
        :param predicate: function which takes the measure results and threshold value; returns test_pass or test_fail
        :param threshold: Numeric or dict to specify the threshold for a given test or collection of tests
        :param hard_fail: Boolean or dict to specify if test failure should result in terminating the current process
        """
        self.description = description
        self.measure = measure
        self.threshold = threshold
        self.predicate = predicate
        self.hard_fail = hard_fail


class CustomSQLTest(Test):
    def __init__(self, description, test, *columns, threshold=None, hard_fail=False):
        """
        :param description:
        :param test:
        :param columns:
        :param threshold:
        :param hard_fail:
        """
        self.description = description
        self.test = test
        self.columns = columns
        self.threshold = threshold
        self.hard_fail = hard_fail


class CSVTest(Test):
    def __init__(self, description, measure, predicate, threshold, hard_fail=False):
        """
        :param description:
        :param measure:
        :param predicate:
        :param threshold:
        :param hard_fail:
        """
        self.description = description
        self.measure = measure
        self.threshold = threshold
        self.predicate = predicate
        self.hard_fail = hard_fail


class TestFactory(object):
    @abc.abstractmethod
    def build_measure(self): pass

    @abc.abstractmethod
    def factory(self): pass


class CustomTestFactory(TestFactory):
    def __init__(self, description, test, *columns, threshold=None, hard_fail=None):
        """
        :param description:
        :param test:
        :param columns:
        :param threshold:
        :param hard_fail:
        """
        self.description = description
        self.test = test
        self.columns = columns
        self.threshold = threshold
        self.hard_fail = hard_fail

    def build_measure(self): pass

    def factory(self):
        return CustomSQLTest(self.description, self.test, *self.columns, threshold=self.threshold,
                             hard_fail=self.hard_fail)


class SQLTestFactory(TestFactory):
    def __init__(
            self,
            description,
            dialect,
            from_,
            predicate,
            threshold,
            *columns,
            where=None,
            hard_fail=False,
            use_ansi=True
    ):
        """
        :param description:
        :param dialect:
        :param from_:
        :param predicate:
        :param threshold:
        :param columns:
        :param where:
        :param hard_fail:
        :param use_ansi:
        """
        self.description = description
        self.threshold = threshold
        self.predicate = predicate
        self.dialect = dialect
        self.from_ = from_
        self.columns = columns
        self.where = where
        self.hard_fail = hard_fail
        self.use_ansi = use_ansi

    @abc.abstractmethod
    def build_measure(self): pass

    def factory(self):
        measure = self.build_measure()
        return SQLTest(self.description, measure, self.predicate, self.threshold, hard_fail=self.hard_fail)


class SQLNullTest(SQLTestFactory):
    def __init__(self, dialect, from_, threshold, *columns, where=None, hard_fail=False, use_ansi=True):
        description = "{column} in table {from_} should have fewer than {threshold} null values."
        predicate = test_predicate_gt
        super().__init__(description, dialect, from_, predicate, threshold, *columns, where=where, hard_fail=hard_fail,
                              use_ansi=use_ansi)

    def build_measure(self):
        measure = SQLNullMeasure(self.dialect, self.from_, *self.columns, where=self.where, use_ansi=self.use_ansi)\
            .factory()
        return measure


class SQLDuplicateTest(SQLTestFactory):
    def __init__(self, dialect, from_, threshold, *columns, where=None, hard_fail=False, use_ansi=True):
        description = "{column} in table {from_} should have fewer than {threshold} duplicate values."
        predicate = test_predicate_gt
        super().__init__(description, dialect, from_, predicate, threshold, *columns, where=where, hard_fail=hard_fail,
                         use_ansi=use_ansi)

    def build_measure(self):
        measure = SQLDuplicateMeasure(self.dialect, self.from_, *self.columns, where=self.where,
                                      use_ansi=self.use_ansi).factory()
        return measure


class SQLSetDuplicateTest(SQLTestFactory):
    def __init__(self, dialect, from_, threshold, *columns, where=None, hard_fail=False, use_ansi=True):
        description = "columns {column} in table {from_} should have fewer than {threshold} duplicate values."
        predicate = test_predicate_gt
        super().__init__(description, dialect, from_, predicate, threshold, *columns, where=where, hard_fail=hard_fail,
                         use_ansi=use_ansi)

    def build_measure(self):
        measure = SQLSetDuplicateMeasure(self.dialect, self.from_, *self.columns, where=self.where,
                                         use_ansi=self.use_ansi).factory()
        return measure


class CSVTestFactory(TestFactory):
    def __init__(self, description, from_, predicate, threshold, *columns, delimiter=',', hard_fail=False):
        """
        :param description:
        :param from_:
        :param predicate:
        :param threshold:
        :param columns:
        :param delimiter:
        :param hard_fail:
        """
        self.description = description
        self.threshold = threshold
        self.predicate = predicate
        self.from_ = from_
        self.columns = columns
        self.delimiter = delimiter
        self.hard_fail = hard_fail

    @abc.abstractmethod
    def build_measure(self): pass

    def factory(self):
        measure = self.build_measure()
        return CSVTest(self.description, measure, self.predicate, self.threshold, hard_fail=self.hard_fail)


class CSVNullTest(CSVTestFactory):
    def __init__(self, from_, threshold, *columns, delimiter=',', hard_fail=False):
        description = "{column} in document {from_} should have fewer than {threshold} null values."
        predicate = test_predicate_gt
        super().__init__(description, from_, predicate, threshold, *columns, delimiter=delimiter, hard_fail=hard_fail)

    def build_measure(self):
        measure = CSVNullMeasure(self.from_, *self.columns, delimiter=self.delimiter).factory()
        return measure


class CSVDuplicateTest(CSVTestFactory):
    def __init__(self, from_, threshold, *columns, delimiter=',', hard_fail=False):
        description = "{column} in document {from_} should have fewer than {threshold} duplicate values."
        predicate = test_predicate_gt
        super().__init__(description, from_, predicate, threshold, *columns, delimiter=delimiter, hard_fail=hard_fail)

    def build_measure(self):
        measure = CSVDuplicateMeasure(self.from_, *self.columns, delimiter=self.delimiter).factory()
        return measure


class CSVSetDuplicateTest(CSVTestFactory):
    def __init__(self, from_, threshold, *columns, delimiter=',', hard_fail=False):
        description = "columns {column} in document {from_} should have fewer than {threshold} duplicate values."
        predicate = test_predicate_gt
        super().__init__(description, from_, predicate, threshold, *columns, delimiter=delimiter, hard_fail=hard_fail)

    def build_measure(self):
        measure = CSVSetDuplicateMeasure(self.from_,  *self.columns, delimiter=self.delimiter).factory()
        return measure





class FetchQueryResults(object):
    def __init__(self, conn, query):
        """
        :param conn:
        :param query:
        """
        self.conn = conn
        self.query = query

        response = self.execute_query()
        self.results = self.fetch_results(response)

    def execute_query(self):
        resposne = self.conn.execute(self.query)
        return resposne

    def fetch_results(self, response):
        result = self.conn.fetch(response)[0]
        result_columns = response.keys()
        query_results = dict(zip(result_columns, result))
        return query_results

    def get_results(self):
        return self.results


import abc

from sqlalchemy.sql import func, distinct

from .sql.helpers import compile_to_dialect
from .sql.measure_logic import measure_proportion_each_column, measure_set_duplication

from .csv.reducers import measure_null, measure_duplicates, measure_set_duplicates


class Measure(object):
    pass


class SQLMeasure(Measure):
    def __init__(self, dialect, from_, query, *columns):
        """
        :param dialect: The SQL dialect for the given database
        :param from_: Schema and table name of table to be tested
        :param query: The SQL query used to calculate measure value
        :param columns: The column names entered as comma separated positional arguments
        """
        self.columns = columns
        self.query = query
        self.from_ = from_
        self.dialect = dialect


class CSVMeasure(Measure):
    def __init__(self, delimiter, from_, reducer, *columns):
        """
        :param delimiter: Separator used in file specified by the `from_` parameter
        :param from_: Path to CSV file
        :param reducer: Function used to calculate measure value
        :param columns: The column names entered as comma separated positional arguments
        """
        self.columns = columns
        self.from_ = from_
        self.reducer = reducer
        self.delimiter = delimiter


class MeasureFactory(object):
    @abc.abstractmethod
    def factory(self): pass


class SQLMeasureFactory(MeasureFactory):
    def __init__(
            self,
            dialect,
            from_,
            *columns,
            where=None,
            use_ansi=True
    ):
        """
        :param dialect: The SQL dialect for the given database
        :param from_: Schema and table name of table to be tested
        :param columns: The column names entered as comma separated positional arguments
        :param where: Conditional logic to be applied to table specified by from_
        :param use_ansi: Boolean to specify if SQL query should be complied to ANSI standards
        """
        self.dialect = dialect
        self.from_ = from_
        self.columns = columns
        self.where = where
        self.use_ansi = use_ansi

    @staticmethod
    def compile_dialect(query, dialect, use_ansi):
        query_ = compile_to_dialect(query, dialect, use_ansi=use_ansi)
        return query_

    @abc.abstractmethod
    def build_measure_query(self):
        pass

    def factory(self):
        query = self.build_measure_query()
        query_ = self.compile_dialect(query, self.dialect, self.use_ansi)
        return SQLMeasure(self.dialect, self.from_, query_, *self.columns)


class SQLNullMeasure(SQLMeasureFactory):
    def __init__(self, dialect, from_, *columns, where=None, use_ansi=True):
        super().__init__(dialect, from_, *columns, where=where, use_ansi=use_ansi)

    def build_measure_query(self):
        aggregate_func = func.count
        measure_query = measure_proportion_each_column(
            self.from_,
            aggregate_func,
            *self.columns,
            where_clause=self.where
        )
        return measure_query


class SQLDuplicateMeasure(SQLMeasureFactory):
    def __init__(self, dialect, from_, *columns, where=None, use_ansi=True):
        super().__init__(dialect, from_, *columns, where=where, use_ansi=use_ansi)

    def build_measure_query(self):
        def aggregate_func(column):
            return func.count(distinct(column))

        measure_query = measure_proportion_each_column(
            self.from_,
            aggregate_func,
            *self.columns,
            where_clause=self.where
        )
        return measure_query


class SQLSetDuplicateMeasure(SQLMeasureFactory):
    def __init__(self, dialect, from_, *columns, where=None, use_ansi=True):
        super().__init__(dialect, from_, *columns, where=where, use_ansi=use_ansi)

    def build_measure_query(self):
        measure_query = measure_set_duplication(self.from_, *self.columns, where_clause=self.where)
        return measure_query


class CSVMeasureFactory(MeasureFactory):
    def __init__(self, from_, *columns, delimiter=','):
        """
        :param from_: Path to CSV file
        :param columns: The column names entered as comma separated positional arguments
        :param delimiter: Separator used in file specified by the `from_` parameter
        """
        self.from_ = from_
        self.columns = columns
        self.delimiter = delimiter

    @abc.abstractmethod
    def build_reducer(self):
        pass

    def factory(self):
        reducer = self.build_reducer()
        return CSVMeasure(self.delimiter, self.from_, reducer, *self.columns)


class CSVNullMeasure(CSVMeasureFactory):
    def __init__(self, from_, *columns, delimiter=','):
        super().__init__(from_, *columns, delimiter=delimiter)

    def build_reducer(self):
        return measure_null


class CSVDuplicateMeasure(CSVMeasureFactory):
    def __init__(self, from_, *columns, delimiter=','):
        super().__init__(from_, *columns, delimiter=delimiter)

    def build_reducer(self):
        return measure_duplicates


class CSVSetDuplicateMeasure(CSVMeasureFactory):
    def __init__(self, from_, *columns, delimiter=','):
        super().__init__(from_, *columns, delimiter=delimiter)

    def build_reducer(self):
        return measure_set_duplicates



import setuptools


with open("README.md", 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name="data-raven",
    version="1.0.0",
    author="Alex Broley",
    author_email="alex.broley@gmail.com",
    description="A Python framework for building data quality tests",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tekkabroley/data-raven",
    packages=setuptools.find_packages(),
    data_files=None,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    python_requires=">=3.6",
    include_package_data=True
)